@@ -28,6 +28,7 @@ pub struct BucketMapHolderStats {
     pub load_disk_missing_us: AtomicU64,
     pub updates_in_mem: AtomicU64,
     pub items: AtomicU64,
+    pub items_us: AtomicU64,
     pub keys: AtomicU64,
     pub deletes: AtomicU64,
     pub inserts: AtomicU64,
@@ -170,10 +171,9 @@ impl BucketMapHolderStats {
         let disk = storage.disk.as_ref();
         let disk_per_bucket_counts = disk
             .map(|disk| {
-                disk.stats
-                    .per_bucket_count
-                    .iter()
-                    .map(|count| count.load(Ordering::Relaxed) as usize)
+                (0..self.bins)
+                    .into_iter()
+                    .map(|i| disk.get_bucket_from_index(i as usize).bucket_len() as usize)
                     .collect::<Vec<_>>()
             })
             .unwrap_or_default();
@@ -323,7 +323,7 @@ impl BucketMapHolderStats {
                 ),
                 (
                     "flush_grow_us",
-                    self.flush_remove_us.swap(0, Ordering::Relaxed),
+                    self.flush_grow_us.swap(0, Ordering::Relaxed),
                     i64
                 ),
                 (
@@ -518,6 +518,7 @@ impl BucketMapHolderStats {
                     i64
                 ),
                 ("items", self.items.swap(0, Ordering::Relaxed), i64),
+                ("items_us", self.items_us.swap(0, Ordering::Relaxed), i64),
                 ("keys", self.keys.swap(0, Ordering::Relaxed), i64),
             );
         }
