@@ -7,7 +7,10 @@
 use {
     crate::{block_cost_limits::*, execute_cost_table::ExecuteCostTable},
     log::*,
-    solana_sdk::{pubkey::Pubkey, transaction::SanitizedTransaction},
+    solana_sdk::{
+        instruction::CompiledInstruction, program_utils::limited_deserialize, pubkey::Pubkey,
+        system_instruction::SystemInstruction, system_program, transaction::SanitizedTransaction,
+    },
     std::collections::HashMap,
 };
 
@@ -20,9 +23,8 @@ pub struct TransactionCost {
     pub signature_cost: u64,
     pub write_lock_cost: u64,
     pub data_bytes_cost: u64,
-    pub builtins_execution_cost: u64,
-    pub bpf_execution_cost: u64,
-    pub is_simple_vote: bool,
+    pub execution_cost: u64,
+    pub account_data_size: u64,
 }
 
 impl Default for TransactionCost {
@@ -32,9 +34,8 @@ impl Default for TransactionCost {
             signature_cost: 0u64,
             write_lock_cost: 0u64,
             data_bytes_cost: 0u64,
-            builtins_execution_cost: 0u64,
-            bpf_execution_cost: 0u64,
-            is_simple_vote: false,
+            execution_cost: 0u64,
+            account_data_size: 0u64,
         }
     }
 }
@@ -52,17 +53,11 @@ impl TransactionCost {
         self.signature_cost = 0;
         self.write_lock_cost = 0;
         self.data_bytes_cost = 0;
-        self.builtins_execution_cost = 0;
-        self.bpf_execution_cost = 0;
-        self.is_simple_vote = false;
+        self.execution_cost = 0;
     }
 
     pub fn sum(&self) -> u64 {
-        self.signature_cost
-            .saturating_add(self.write_lock_cost)
-            .saturating_add(self.data_bytes_cost)
-            .saturating_add(self.builtins_execution_cost)
-            .saturating_add(self.bpf_execution_cost)
+        self.signature_cost + self.write_lock_cost + self.data_bytes_cost + self.execution_cost
     }
 }
 
@@ -82,6 +77,7 @@ impl CostModel {
         cost_table
             .iter()
             .map(|(key, cost)| (key, cost))
+            .chain(BUILT_IN_INSTRUCTION_COSTS.iter())
             .for_each(|(program_id, cost)| {
                 match self
                     .instruction_execution_cost_table
@@ -113,10 +109,8 @@ impl CostModel {
         tx_cost.signature_cost = self.get_signature_cost(transaction);
         self.get_write_lock_cost(&mut tx_cost, transaction);
         tx_cost.data_bytes_cost = self.get_data_bytes_cost(transaction);
-        let (builtins_cost, bpf_cost) = self.get_transaction_cost(transaction);
-        tx_cost.builtins_execution_cost = builtins_cost;
-        tx_cost.bpf_execution_cost = bpf_cost;
-        tx_cost.is_simple_vote = transaction.is_simple_vote_transaction();
+        tx_cost.execution_cost = self.get_transaction_cost(transaction);
+        tx_cost.account_data_size = self.calculate_account_data_size(transaction);
 
         debug!("transaction {:?} has cost {:?}", transaction, tx_cost);
         tx_cost
@@ -153,10 +147,6 @@ impl CostModel {
         }
     }
 
-    pub fn get_program_keys(&self) -> Vec<&Pubkey> {
-        self.instruction_execution_cost_table.get_program_keys()
-    }
-
     fn get_signature_cost(&self, transaction: &SanitizedTransaction) -> u64 {
         transaction.signatures().len() as u64 * SIGNATURE_COST
     }
@@ -167,14 +157,18 @@ impl CostModel {
         transaction: &SanitizedTransaction,
     ) {
         let message = transaction.message();
-        message.account_keys_iter().enumerate().for_each(|(i, k)| {
-            let is_writable = message.is_writable(i);
+        message
+            .account_keys()
+            .iter()
+            .enumerate()
+            .for_each(|(i, k)| {
+                let is_writable = message.is_writable(i);
 
-            if is_writable {
-                tx_cost.writable_accounts.push(*k);
-                tx_cost.write_lock_cost += WRITE_LOCK_UNITS;
-            }
-        });
+                if is_writable {
+                    tx_cost.writable_accounts.push(*k);
+                    tx_cost.write_lock_cost += WRITE_LOCK_UNITS;
+                }
+            });
     }
 
     fn get_data_bytes_cost(&self, transaction: &SanitizedTransaction) -> u64 {
@@ -188,25 +182,72 @@ impl CostModel {
         data_bytes_cost
     }
 
-    fn get_transaction_cost(&self, transaction: &SanitizedTransaction) -> (u64, u64) {
-        let mut builtin_costs = 0u64;
-        let mut bpf_costs = 0u64;
+    fn get_transaction_cost(&self, transaction: &SanitizedTransaction) -> u64 {
+        let mut cost: u64 = 0;
 
         for (program_id, instruction) in transaction.message().program_instructions_iter() {
-            // to keep the same behavior, look for builtin first
-            if let Some(builtin_cost) = BUILT_IN_INSTRUCTION_COSTS.get(program_id) {
-                builtin_costs = builtin_costs.saturating_add(*builtin_cost);
-            } else {
-                let instruction_cost = self.find_instruction_cost(program_id);
-                trace!(
-                    "instruction {:?} has cost of {}",
+            let instruction_cost = self.find_instruction_cost(program_id);
+            trace!(
+                "instruction {:?} has cost of {}",
+                instruction,
+                instruction_cost
+            );
+            cost = cost.saturating_add(instruction_cost);
+        }
+        cost
+    }
+
+    fn calculate_account_data_size_on_deserialized_system_instruction(
+        instruction: SystemInstruction,
+    ) -> u64 {
+        match instruction {
+            SystemInstruction::CreateAccount {
+                lamports: _lamports,
+                space,
+                owner: _owner,
+            } => space,
+            SystemInstruction::CreateAccountWithSeed {
+                base: _base,
+                seed: _seed,
+                lamports: _lamports,
+                space,
+                owner: _owner,
+            } => space,
+            SystemInstruction::Allocate { space } => space,
+            SystemInstruction::AllocateWithSeed {
+                base: _base,
+                seed: _seed,
+                space,
+                owner: _owner,
+            } => space,
+            _ => 0,
+        }
+    }
+
+    fn calculate_account_data_size_on_instruction(
+        program_id: &Pubkey,
+        instruction: &CompiledInstruction,
+    ) -> u64 {
+        if program_id == &system_program::id() {
+            if let Ok(instruction) = limited_deserialize(&instruction.data) {
+                return Self::calculate_account_data_size_on_deserialized_system_instruction(
                     instruction,
-                    instruction_cost
                 );
-                bpf_costs = bpf_costs.saturating_add(instruction_cost);
             }
         }
-        (builtin_costs, bpf_costs)
+        0
+    }
+
+    /// eventually, potentially determine account data size of all writable accounts
+    /// at the moment, calculate account data size of account creation
+    fn calculate_account_data_size(&self, transaction: &SanitizedTransaction) -> u64 {
+        transaction
+            .message()
+            .program_instructions_iter()
+            .map(|(program_id, instruction)| {
+                Self::calculate_account_data_size_on_instruction(program_id, instruction)
+            })
+            .sum()
     }
 }
 
@@ -271,32 +312,50 @@ mod tests {
     }
 
     #[test]
-    fn test_iterating_instruction_cost_by_program_keys() {
-        solana_logger::setup();
-        let mut testee = CostModel::default();
-
-        let mut test_key_and_cost = HashMap::<Pubkey, u64>::new();
-        (0u64..10u64).for_each(|n| {
-            test_key_and_cost.insert(Pubkey::new_unique(), n);
-        });
-
-        test_key_and_cost.iter().for_each(|(key, cost)| {
-            let _ = testee.upsert_instruction_cost(key, *cost).unwrap();
-            info!("key {:?} cost {}", key, cost);
-        });
-
-        let keys = testee.get_program_keys();
-        // verify each key has pre-set value
-        keys.iter().for_each(|key| {
-            let expected_cost = test_key_and_cost.get(key).unwrap();
-            info!(
-                "check key {:?} expect {} find {}",
-                key,
-                expected_cost,
-                testee.find_instruction_cost(key)
+    fn test_cost_model_data_len_cost() {
+        let lamports = 0;
+        let owner = Pubkey::default();
+        let seed = String::default();
+        let space = 100;
+        let base = Pubkey::default();
+        for instruction in [
+            SystemInstruction::CreateAccount {
+                lamports,
+                space,
+                owner,
+            },
+            SystemInstruction::CreateAccountWithSeed {
+                base,
+                seed: seed.clone(),
+                lamports,
+                space,
+                owner,
+            },
+            SystemInstruction::Allocate { space },
+            SystemInstruction::AllocateWithSeed {
+                base,
+                seed,
+                space,
+                owner,
+            },
+        ] {
+            assert_eq!(
+                space,
+                CostModel::calculate_account_data_size_on_deserialized_system_instruction(
+                    instruction
+                )
             );
-            assert_eq!(*expected_cost, testee.find_instruction_cost(key));
-        });
+        }
+        assert_eq!(
+            0,
+            CostModel::calculate_account_data_size_on_deserialized_system_instruction(
+                SystemInstruction::TransferWithSeed {
+                    lamports,
+                    from_seed: String::default(),
+                    from_owner: Pubkey::default(),
+                }
+            )
+        );
     }
 
     #[test]
@@ -313,13 +372,14 @@ mod tests {
         );
 
         // expected cost for one system transfer instructions
-        let expected_execution_cost = BUILT_IN_INSTRUCTION_COSTS
-            .get(&system_program::id())
-            .unwrap();
+        let expected_cost = 8;
 
-        let testee = CostModel::default();
+        let mut testee = CostModel::default();
+        testee
+            .upsert_instruction_cost(&system_program::id(), expected_cost)
+            .unwrap();
         assert_eq!(
-            (*expected_execution_cost, 0),
+            expected_cost,
             testee.get_transaction_cost(&simple_transaction)
         );
     }
@@ -341,13 +401,14 @@ mod tests {
         debug!("many transfer transaction {:?}", tx);
 
         // expected cost for two system transfer instructions
-        let program_cost = BUILT_IN_INSTRUCTION_COSTS
-            .get(&system_program::id())
-            .unwrap();
+        let program_cost = 8;
         let expected_cost = program_cost * 2;
 
-        let testee = CostModel::default();
-        assert_eq!((expected_cost, 0), testee.get_transaction_cost(&tx));
+        let mut testee = CostModel::default();
+        testee
+            .upsert_instruction_cost(&system_program::id(), program_cost)
+            .unwrap();
+        assert_eq!(expected_cost, testee.get_transaction_cost(&tx));
     }
 
     #[test]
@@ -379,7 +440,7 @@ mod tests {
 
         // expected cost for two random/unknown program is
         let expected_cost = testee.instruction_execution_cost_table.get_mode() * 2;
-        assert_eq!((0, expected_cost), result);
+        assert_eq!(expected_cost, result);
     }
 
     #[test]
@@ -444,14 +505,15 @@ mod tests {
         ));
 
         let expected_account_cost = WRITE_LOCK_UNITS * 2;
-        let expected_execution_cost = BUILT_IN_INSTRUCTION_COSTS
-            .get(&system_program::id())
-            .unwrap();
+        let expected_execution_cost = 8;
 
-        let cost_model = CostModel::default();
+        let mut cost_model = CostModel::default();
+        cost_model
+            .upsert_instruction_cost(&system_program::id(), expected_execution_cost)
+            .unwrap();
         let tx_cost = cost_model.calculate_cost(&tx);
         assert_eq!(expected_account_cost, tx_cost.write_lock_cost);
-        assert_eq!(*expected_execution_cost, tx_cost.builtins_execution_cost);
+        assert_eq!(expected_execution_cost, tx_cost.execution_cost);
         assert_eq!(2, tx_cost.writable_accounts.len());
     }
 
@@ -548,14 +610,14 @@ mod tests {
             assert_eq!(*cost, cost_model.find_instruction_cost(id));
         }
 
-        // verify built-in programs are not in bpf_costs
+        // verify built-in programs
         assert!(cost_model
             .instruction_execution_cost_table
             .get_cost(&system_program::id())
-            .is_none());
+            .is_some());
         assert!(cost_model
             .instruction_execution_cost_table
             .get_cost(&solana_vote_program::id())
-            .is_none());
+            .is_some());
     }
 }
