@@ -3,11 +3,7 @@ use {
     solana_measure::measure::Measure,
     solana_program_runtime::{
         compute_budget::ComputeBudget,
-        instruction_recorder::InstructionRecorder,
-        invoke_context::{
-            BuiltinProgram, Executors, InvokeContext, ProcessInstructionResult,
-            TransactionAccountRefCell,
-        },
+        invoke_context::{BuiltinProgram, Executors, InvokeContext},
         log_collector::LogCollector,
         sysvar_cache::SysvarCache,
         timings::ExecuteTimings,
@@ -22,6 +18,7 @@ use {
         saturating_add_assign,
         sysvar::instructions,
         transaction::TransactionError,
+        transaction_context::{InstructionAccount, TransactionContext},
     },
     std::{borrow::Cow, cell::RefCell, rc::Rc, sync::Arc},
 };
@@ -38,9 +35,6 @@ impl ::solana_frozen_abi::abi_example::AbiExample for MessageProcessor {
     }
 }
 
-/// Trace of all instructions attempted
-pub type InstructionTrace = Vec<InstructionRecorder>;
-
 /// Resultant information gathered from calling process_message()
 #[derive(Debug, Default, Clone, Copy, PartialEq, Eq)]
 pub struct ProcessedMessageInfo {
@@ -59,11 +53,10 @@ impl MessageProcessor {
         builtin_programs: &[BuiltinProgram],
         message: &SanitizedMessage,
         program_indices: &[Vec<usize>],
-        accounts: &[TransactionAccountRefCell],
+        transaction_context: &mut TransactionContext,
         rent: Rent,
         log_collector: Option<Rc<RefCell<LogCollector>>>,
         executors: Rc<RefCell<Executors>>,
-        instruction_trace: &mut InstructionTrace,
         feature_set: Arc<FeatureSet>,
         compute_budget: ComputeBudget,
         timings: &mut ExecuteTimings,
@@ -71,11 +64,10 @@ impl MessageProcessor {
         blockhash: Hash,
         lamports_per_signature: u64,
         current_accounts_data_len: u64,
-        accumulated_consumed_units: &mut u64,
     ) -> Result<ProcessedMessageInfo, TransactionError> {
         let mut invoke_context = InvokeContext::new(
+            transaction_context,
             rent,
-            accounts,
             builtin_programs,
             Cow::Borrowed(sysvar_cache),
             log_collector,
@@ -93,12 +85,6 @@ impl MessageProcessor {
             .zip(program_indices.iter())
             .enumerate()
         {
-            invoke_context.record_top_level_instruction(
-                instruction.decompile(message).map_err(|err| {
-                    TransactionError::InstructionError(instruction_index as u8, err)
-                })?,
-            );
-
             if invoke_context
                 .feature_set
                 .is_active(&prevent_calling_precompiles_as_programs::id())
@@ -110,32 +96,44 @@ impl MessageProcessor {
 
             // Fixup the special instructions key if present
             // before the account pre-values are taken care of
-            for (pubkey, account) in accounts.iter().take(message.account_keys_len()) {
-                if instructions::check_id(pubkey) {
-                    let mut mut_account_ref = account.borrow_mut();
-                    instructions::store_current_index(
-                        mut_account_ref.data_as_mut_slice(),
-                        instruction_index as u16,
-                    );
-                    break;
-                }
+            if let Some(account_index) = invoke_context
+                .transaction_context
+                .find_index_of_account(&instructions::id())
+            {
+                let mut mut_account_ref = invoke_context
+                    .transaction_context
+                    .get_account_at_index(account_index)
+                    .map_err(|_| TransactionError::InvalidAccountIndex)?
+                    .borrow_mut();
+                instructions::store_current_index(
+                    mut_account_ref.data_as_mut_slice(),
+                    instruction_index as u16,
+                );
             }
 
+            let instruction_accounts = instruction
+                .accounts
+                .iter()
+                .map(|index_in_transaction| {
+                    let index_in_transaction = *index_in_transaction as usize;
+                    InstructionAccount {
+                        index_in_transaction,
+                        index_in_caller: program_indices.len().saturating_add(index_in_transaction),
+                        is_signer: message.is_signer(index_in_transaction),
+                        is_writable: message.is_writable(index_in_transaction),
+                    }
+                })
+                .collect::<Vec<_>>();
             let mut time = Measure::start("execute_instruction");
-            let ProcessInstructionResult {
-                compute_units_consumed,
-                result,
-            } = invoke_context.process_instruction(
-                message,
-                instruction,
+            let mut compute_units_consumed = 0;
+            let result = invoke_context.process_instruction(
+                &instruction.data,
+                &instruction_accounts,
                 program_indices,
-                &[],
-                &[],
+                &mut compute_units_consumed,
                 timings,
             );
             time.stop();
-            *accumulated_consumed_units =
-                accumulated_consumed_units.saturating_add(compute_units_consumed);
             timings.details.accumulate_program(
                 program_id,
                 time.as_us(),
@@ -147,13 +145,9 @@ impl MessageProcessor {
                 timings.execute_accessories.process_instructions.total_us,
                 time.as_us()
             );
-
-            result.map_err(|err| {
-                instruction_trace.append(invoke_context.get_instruction_trace_mut());
-                TransactionError::InstructionError(instruction_index as u8, err)
-            })?;
+            result
+                .map_err(|err| TransactionError::InstructionError(instruction_index as u8, err))?;
         }
-        instruction_trace.append(invoke_context.get_instruction_trace_mut());
         Ok(ProcessedMessageInfo {
             accounts_data_len_delta: invoke_context.get_accounts_data_meter().delta(),
         })
@@ -168,7 +162,6 @@ mod tests {
         solana_sdk::{
             account::{AccountSharedData, ReadableAccount},
             instruction::{AccountMeta, Instruction, InstructionError},
-            keyed_account::keyed_account_at_index,
             message::Message,
             native_loader::{self, create_loadable_account_for_test},
             pubkey::Pubkey,
@@ -191,36 +184,33 @@ mod tests {
         #[derive(Serialize, Deserialize)]
         enum MockSystemInstruction {
             Correct,
-            AttemptCredit { lamports: u64 },
-            AttemptDataChange { data: u8 },
+            TransferLamports { lamports: u64 },
+            ChangeData { data: u8 },
         }
 
         fn mock_system_process_instruction(
-            first_instruction_account: usize,
+            _first_instruction_account: usize,
             data: &[u8],
             invoke_context: &mut InvokeContext,
         ) -> Result<(), InstructionError> {
-            let keyed_accounts = invoke_context.get_keyed_accounts()?;
+            let transaction_context = &invoke_context.transaction_context;
+            let instruction_context = transaction_context.get_current_instruction_context()?;
             if let Ok(instruction) = bincode::deserialize(data) {
                 match instruction {
                     MockSystemInstruction::Correct => Ok(()),
-                    MockSystemInstruction::AttemptCredit { lamports } => {
-                        keyed_account_at_index(keyed_accounts, first_instruction_account)?
-                            .account
-                            .borrow_mut()
+                    MockSystemInstruction::TransferLamports { lamports } => {
+                        instruction_context
+                            .try_borrow_instruction_account(transaction_context, 0)?
                             .checked_sub_lamports(lamports)?;
-                        keyed_account_at_index(keyed_accounts, first_instruction_account + 1)?
-                            .account
-                            .borrow_mut()
+                        instruction_context
+                            .try_borrow_instruction_account(transaction_context, 1)?
                             .checked_add_lamports(lamports)?;
                         Ok(())
                     }
-                    // Change data in a read-only account
-                    MockSystemInstruction::AttemptDataChange { data } => {
-                        keyed_account_at_index(keyed_accounts, first_instruction_account + 1)?
-                            .account
-                            .borrow_mut()
-                            .set_data(vec![data]);
+                    MockSystemInstruction::ChangeData { data } => {
+                        instruction_context
+                            .try_borrow_instruction_account(transaction_context, 1)?
+                            .set_data(&[data]);
                         Ok(())
                     }
                 }
@@ -236,46 +226,51 @@ mod tests {
             process_instruction: mock_system_process_instruction,
         }];
 
-        let program_account = Rc::new(RefCell::new(create_loadable_account_for_test(
-            "mock_system_program",
-        )));
         let accounts = vec![
             (
                 solana_sdk::pubkey::new_rand(),
-                AccountSharedData::new_ref(100, 1, &mock_system_program_id),
+                AccountSharedData::new(100, 1, &mock_system_program_id),
             ),
             (
                 solana_sdk::pubkey::new_rand(),
-                AccountSharedData::new_ref(0, 1, &mock_system_program_id),
+                AccountSharedData::new(0, 1, &mock_system_program_id),
+            ),
+            (
+                mock_system_program_id,
+                create_loadable_account_for_test("mock_system_program"),
             ),
-            (mock_system_program_id, program_account),
         ];
+        let mut transaction_context = TransactionContext::new(accounts, 1, 3);
         let program_indices = vec![vec![2]];
-
         let executors = Rc::new(RefCell::new(Executors::default()));
-
         let account_metas = vec![
-            AccountMeta::new(accounts[0].0, true),
-            AccountMeta::new_readonly(accounts[1].0, false),
+            AccountMeta::new(
+                *transaction_context.get_key_of_account_at_index(0).unwrap(),
+                true,
+            ),
+            AccountMeta::new_readonly(
+                *transaction_context.get_key_of_account_at_index(1).unwrap(),
+                false,
+            ),
         ];
+
         let message = SanitizedMessage::Legacy(Message::new(
             &[Instruction::new_with_bincode(
                 mock_system_program_id,
                 &MockSystemInstruction::Correct,
                 account_metas.clone(),
             )],
-            Some(&accounts[0].0),
+            Some(transaction_context.get_key_of_account_at_index(0).unwrap()),
         ));
         let sysvar_cache = SysvarCache::default();
         let result = MessageProcessor::process_message(
             builtin_programs,
             &message,
             &program_indices,
-            &accounts,
+            &mut transaction_context,
             rent_collector.rent,
             None,
             executors.clone(),
-            &mut Vec::new(),
             Arc::new(FeatureSet::all_enabled()),
             ComputeBudget::default(),
             &mut ExecuteTimings::default(),
@@ -283,30 +278,41 @@ mod tests {
             Hash::default(),
             0,
             0,
-            &mut 0,
         );
         assert!(result.is_ok());
-        assert_eq!(accounts[0].1.borrow().lamports(), 100);
-        assert_eq!(accounts[1].1.borrow().lamports(), 0);
+        assert_eq!(
+            transaction_context
+                .get_account_at_index(0)
+                .unwrap()
+                .borrow()
+                .lamports(),
+            100
+        );
+        assert_eq!(
+            transaction_context
+                .get_account_at_index(1)
+                .unwrap()
+                .borrow()
+                .lamports(),
+            0
+        );
 
         let message = SanitizedMessage::Legacy(Message::new(
             &[Instruction::new_with_bincode(
                 mock_system_program_id,
-                &MockSystemInstruction::AttemptCredit { lamports: 50 },
+                &MockSystemInstruction::TransferLamports { lamports: 50 },
                 account_metas.clone(),
             )],
-            Some(&accounts[0].0),
+            Some(transaction_context.get_key_of_account_at_index(0).unwrap()),
         ));
-
         let result = MessageProcessor::process_message(
             builtin_programs,
             &message,
             &program_indices,
-            &accounts,
+            &mut transaction_context,
             rent_collector.rent,
             None,
             executors.clone(),
-            &mut Vec::new(),
             Arc::new(FeatureSet::all_enabled()),
             ComputeBudget::default(),
             &mut ExecuteTimings::default(),
@@ -314,7 +320,6 @@ mod tests {
             Hash::default(),
             0,
             0,
-            &mut 0,
         );
         assert_eq!(
             result,
@@ -327,21 +332,19 @@ mod tests {
         let message = SanitizedMessage::Legacy(Message::new(
             &[Instruction::new_with_bincode(
                 mock_system_program_id,
-                &MockSystemInstruction::AttemptDataChange { data: 50 },
+                &MockSystemInstruction::ChangeData { data: 50 },
                 account_metas,
             )],
-            Some(&accounts[0].0),
+            Some(transaction_context.get_key_of_account_at_index(0).unwrap()),
         ));
-
         let result = MessageProcessor::process_message(
             builtin_programs,
             &message,
             &program_indices,
-            &accounts,
+            &mut transaction_context,
             rent_collector.rent,
             None,
             executors,
-            &mut Vec::new(),
             Arc::new(FeatureSet::all_enabled()),
             ComputeBudget::default(),
             &mut ExecuteTimings::default(),
@@ -349,7 +352,6 @@ mod tests {
             Hash::default(),
             0,
             0,
-            &mut 0,
         );
         assert_eq!(
             result,
@@ -370,67 +372,49 @@ mod tests {
         }
 
         fn mock_system_process_instruction(
-            first_instruction_account: usize,
+            _first_instruction_account: usize,
             data: &[u8],
             invoke_context: &mut InvokeContext,
         ) -> Result<(), InstructionError> {
-            let keyed_accounts = invoke_context.get_keyed_accounts()?;
+            let transaction_context = &invoke_context.transaction_context;
+            let instruction_context = transaction_context.get_current_instruction_context()?;
+            let mut to_account =
+                instruction_context.try_borrow_instruction_account(transaction_context, 1)?;
             if let Ok(instruction) = bincode::deserialize(data) {
                 match instruction {
                     MockSystemInstruction::BorrowFail => {
-                        let from_account =
-                            keyed_account_at_index(keyed_accounts, first_instruction_account)?
-                                .try_account_ref_mut()?;
-                        let dup_account =
-                            keyed_account_at_index(keyed_accounts, first_instruction_account + 2)?
-                                .try_account_ref_mut()?;
-                        if from_account.lamports() != dup_account.lamports() {
+                        let from_account = instruction_context
+                            .try_borrow_instruction_account(transaction_context, 0)?;
+                        let dup_account = instruction_context
+                            .try_borrow_instruction_account(transaction_context, 2)?;
+                        if from_account.get_lamports() != dup_account.get_lamports() {
                             return Err(InstructionError::InvalidArgument);
                         }
                         Ok(())
                     }
                     MockSystemInstruction::MultiBorrowMut => {
-                        let from_lamports = {
-                            let from_account =
-                                keyed_account_at_index(keyed_accounts, first_instruction_account)?
-                                    .try_account_ref_mut()?;
-                            from_account.lamports()
-                        };
-                        let dup_lamports = {
-                            let dup_account = keyed_account_at_index(
-                                keyed_accounts,
-                                first_instruction_account + 2,
-                            )?
-                            .try_account_ref_mut()?;
-                            dup_account.lamports()
-                        };
-                        if from_lamports != dup_lamports {
+                        let lamports_a = instruction_context
+                            .try_borrow_instruction_account(transaction_context, 0)?
+                            .get_lamports();
+                        let lamports_b = instruction_context
+                            .try_borrow_instruction_account(transaction_context, 2)?
+                            .get_lamports();
+                        if lamports_a != lamports_b {
                             return Err(InstructionError::InvalidArgument);
                         }
                         Ok(())
                     }
                     MockSystemInstruction::DoWork { lamports, data } => {
-                        {
-                            let mut to_account = keyed_account_at_index(
-                                keyed_accounts,
-                                first_instruction_account + 1,
-                            )?
-                            .try_account_ref_mut()?;
-                            let mut dup_account = keyed_account_at_index(
-                                keyed_accounts,
-                                first_instruction_account + 2,
-                            )?
-                            .try_account_ref_mut()?;
-                            dup_account.checked_sub_lamports(lamports)?;
-                            to_account.checked_add_lamports(lamports)?;
-                            dup_account.set_data(vec![data]);
-                        }
-                        keyed_account_at_index(keyed_accounts, first_instruction_account)?
-                            .try_account_ref_mut()?
-                            .checked_sub_lamports(lamports)?;
-                        keyed_account_at_index(keyed_accounts, first_instruction_account + 1)?
-                            .try_account_ref_mut()?
-                            .checked_add_lamports(lamports)?;
+                        let mut dup_account = instruction_context
+                            .try_borrow_instruction_account(transaction_context, 2)?;
+                        dup_account.checked_sub_lamports(lamports)?;
+                        to_account.checked_add_lamports(lamports)?;
+                        dup_account.set_data(&[data]);
+                        drop(dup_account);
+                        let mut from_account = instruction_context
+                            .try_borrow_instruction_account(transaction_context, 0)?;
+                        from_account.checked_sub_lamports(lamports)?;
+                        to_account.checked_add_lamports(lamports)?;
                         Ok(())
                     }
                 }
@@ -446,28 +430,36 @@ mod tests {
             process_instruction: mock_system_process_instruction,
         }];
 
-        let program_account = Rc::new(RefCell::new(create_loadable_account_for_test(
-            "mock_system_program",
-        )));
         let accounts = vec![
             (
                 solana_sdk::pubkey::new_rand(),
-                AccountSharedData::new_ref(100, 1, &mock_program_id),
+                AccountSharedData::new(100, 1, &mock_program_id),
             ),
             (
                 solana_sdk::pubkey::new_rand(),
-                AccountSharedData::new_ref(0, 1, &mock_program_id),
+                AccountSharedData::new(0, 1, &mock_program_id),
+            ),
+            (
+                mock_program_id,
+                create_loadable_account_for_test("mock_system_program"),
             ),
-            (mock_program_id, program_account),
         ];
+        let mut transaction_context = TransactionContext::new(accounts, 1, 3);
         let program_indices = vec![vec![2]];
-
         let executors = Rc::new(RefCell::new(Executors::default()));
-
         let account_metas = vec![
-            AccountMeta::new(accounts[0].0, true),
-            AccountMeta::new(accounts[1].0, false),
-            AccountMeta::new(accounts[0].0, false),
+            AccountMeta::new(
+                *transaction_context.get_key_of_account_at_index(0).unwrap(),
+                true,
+            ),
+            AccountMeta::new(
+                *transaction_context.get_key_of_account_at_index(1).unwrap(),
+                false,
+            ),
+            AccountMeta::new(
+                *transaction_context.get_key_of_account_at_index(0).unwrap(),
+                false,
+            ),
         ];
 
         // Try to borrow mut the same account
@@ -477,18 +469,17 @@ mod tests {
                 &MockSystemInstruction::BorrowFail,
                 account_metas.clone(),
             )],
-            Some(&accounts[0].0),
+            Some(transaction_context.get_key_of_account_at_index(0).unwrap()),
         ));
         let sysvar_cache = SysvarCache::default();
         let result = MessageProcessor::process_message(
             builtin_programs,
             &message,
             &program_indices,
-            &accounts,
+            &mut transaction_context,
             rent_collector.rent,
             None,
             executors.clone(),
-            &mut Vec::new(),
             Arc::new(FeatureSet::all_enabled()),
             ComputeBudget::default(),
             &mut ExecuteTimings::default(),
@@ -496,7 +487,6 @@ mod tests {
             Hash::default(),
             0,
             0,
-            &mut 0,
         );
         assert_eq!(
             result,
@@ -513,17 +503,16 @@ mod tests {
                 &MockSystemInstruction::MultiBorrowMut,
                 account_metas.clone(),
             )],
-            Some(&accounts[0].0),
+            Some(transaction_context.get_key_of_account_at_index(0).unwrap()),
         ));
         let result = MessageProcessor::process_message(
             builtin_programs,
             &message,
             &program_indices,
-            &accounts,
+            &mut transaction_context,
             rent_collector.rent,
             None,
             executors.clone(),
-            &mut Vec::new(),
             Arc::new(FeatureSet::all_enabled()),
             ComputeBudget::default(),
             &mut ExecuteTimings::default(),
@@ -531,11 +520,10 @@ mod tests {
             Hash::default(),
             0,
             0,
-            &mut 0,
         );
         assert!(result.is_ok());
 
-        // Do work on the same account but at different location in keyed_accounts[]
+        // Do work on the same transaction account but at different instruction accounts
         let message = SanitizedMessage::Legacy(Message::new(
             &[Instruction::new_with_bincode(
                 mock_program_id,
@@ -545,17 +533,16 @@ mod tests {
                 },
                 account_metas,
             )],
-            Some(&accounts[0].0),
+            Some(transaction_context.get_key_of_account_at_index(0).unwrap()),
         ));
         let result = MessageProcessor::process_message(
             builtin_programs,
             &message,
             &program_indices,
-            &accounts,
+            &mut transaction_context,
             rent_collector.rent,
             None,
             executors,
-            &mut Vec::new(),
             Arc::new(FeatureSet::all_enabled()),
             ComputeBudget::default(),
             &mut ExecuteTimings::default(),
@@ -563,12 +550,32 @@ mod tests {
             Hash::default(),
             0,
             0,
-            &mut 0,
         );
         assert!(result.is_ok());
-        assert_eq!(accounts[0].1.borrow().lamports(), 80);
-        assert_eq!(accounts[1].1.borrow().lamports(), 20);
-        assert_eq!(accounts[0].1.borrow().data(), &vec![42]);
+        assert_eq!(
+            transaction_context
+                .get_account_at_index(0)
+                .unwrap()
+                .borrow()
+                .lamports(),
+            80
+        );
+        assert_eq!(
+            transaction_context
+                .get_account_at_index(1)
+                .unwrap()
+                .borrow()
+                .lamports(),
+            20
+        );
+        assert_eq!(
+            transaction_context
+                .get_account_at_index(0)
+                .unwrap()
+                .borrow()
+                .data(),
+            &vec![42]
+        );
     }
 
     #[test]
@@ -586,14 +593,15 @@ mod tests {
             process_instruction: mock_process_instruction,
         }];
 
-        let secp256k1_account = AccountSharedData::new_ref(1, 0, &native_loader::id());
-        secp256k1_account.borrow_mut().set_executable(true);
-        let mock_program_account = AccountSharedData::new_ref(1, 0, &native_loader::id());
-        mock_program_account.borrow_mut().set_executable(true);
+        let mut secp256k1_account = AccountSharedData::new(1, 0, &native_loader::id());
+        secp256k1_account.set_executable(true);
+        let mut mock_program_account = AccountSharedData::new(1, 0, &native_loader::id());
+        mock_program_account.set_executable(true);
         let accounts = vec![
             (secp256k1_program::id(), secp256k1_account),
             (mock_program_id, mock_program_account),
         ];
+        let mut transaction_context = TransactionContext::new(accounts, 1, 1);
 
         let message = SanitizedMessage::Legacy(Message::new(
             &[
@@ -610,11 +618,10 @@ mod tests {
             builtin_programs,
             &message,
             &[vec![0], vec![1]],
-            &accounts,
+            &mut transaction_context,
             RentCollector::default().rent,
             None,
             Rc::new(RefCell::new(Executors::default())),
-            &mut Vec::new(),
             Arc::new(FeatureSet::all_enabled()),
             ComputeBudget::default(),
             &mut ExecuteTimings::default(),
@@ -622,9 +629,7 @@ mod tests {
             Hash::default(),
             0,
             0,
-            &mut 0,
         );
-
         assert_eq!(
             result,
             Err(TransactionError::InstructionError(
